package dev.fitko.fitconnect.core.http;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import okhttp3.Headers;
import okhttp3.Interceptor;
import okhttp3.MediaType;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.ResponseBody;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.net.Proxy;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Map;

public class HttpClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(HttpClient.class);

    private final OkHttpClient httpClient;
    private final ObjectMapper objectMapper = new ObjectMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    private final boolean throwExceptionsOnFailure;

    public HttpClient() {
        this(true);
    }

    public HttpClient(boolean throwExceptionsOnFailure) {
        this(throwExceptionsOnFailure, List.of());
    }

    public HttpClient(List<Interceptor> interceptors) {
        this(true, interceptors);
    }

    public HttpClient(boolean throwExceptionsOnFailure, List<Interceptor> interceptors) {
        LOGGER.info("Creating HttpClient without proxy configuration.");
        this.httpClient = builderWithInterceptors(interceptors).build();
        this.throwExceptionsOnFailure = throwExceptionsOnFailure;
    }

    public HttpClient(boolean throwExceptionsOnFailure, List<Interceptor> interceptors, Proxy proxy) {
        LOGGER.info("Creating HttpClient with proxy configuration: {}", proxy);
        this.httpClient = builderWithInterceptors(interceptors).proxy(proxy).build();
        this.throwExceptionsOnFailure = throwExceptionsOnFailure;
    }

    private OkHttpClient.Builder builderWithInterceptors(List<Interceptor> interceptors) {

        OkHttpClient.Builder builder = new OkHttpClient.Builder();
        interceptors.forEach(builder::addInterceptor);

        return builder;
    }

    public <R> HttpResponse<R> get(String url, Class<R> responseType) {
        return get(url, Map.of(), responseType);
    }

    public <R> HttpResponse<R> get(String url, Map<String, String> headers, Class<R> responseType) {

        Request request = new Request.Builder()
                .url(url)
                .get()
                .headers(Headers.of(headers))
                .build();

        try (Response response = this.httpClient.newCall(request).execute()) {
            return this.evaluateStatusAndRespond(response, responseType);
        } catch (IOException exception) {
            throw new RestApiException("HTTP GET call to '" + url + "' failed.", exception);
        }
    }

    public <P, R> HttpResponse<R> post(String url, P httpPayload, Class<R> responseType) {
        return post(url, Map.of(), httpPayload, responseType);
    }

    public <P, R> HttpResponse<R> post(String url, Map<String, String> headers, P httpPayload, Class<R> responseType) {

        try {

            RequestBody requestBody = this.createRequestBody(headers, httpPayload);
            Request request = new Request.Builder()
                    .url(url)
                    .post(requestBody)
                    .headers(Headers.of(headers))
                    .build();

            try (Response response = this.httpClient.newCall(request).execute()) {
                return this.evaluateStatusAndRespond(response, responseType);
            }

        } catch (IOException exception) {
            throw new RestApiException("HTTP POST call to '" + url + "' failed.", exception);
        }
    }

    public <P, R> HttpResponse<R> put(String url, P httpPayload, Class<R> responseType) {
        return put(url, Map.of(), httpPayload, responseType);
    }

    public <P, R> HttpResponse<R> put(String url, Map<String, String> headers, P httpPayload, Class<R> responseType) {

        try {

            RequestBody requestBody = this.createRequestBody(headers, httpPayload);
            Request request = new Request.Builder()
                    .url(url)
                    .put(requestBody)
                    .headers(Headers.of(headers))
                    .build();

            try (Response response = this.httpClient.newCall(request).execute()) {
                return this.evaluateStatusAndRespond(response, responseType);
            }

        } catch (IOException exception) {
            throw new RestApiException("HTTP PUT call to '" + url + "' failed.", exception);
        }
    }

    /**
     * The methods {@link ObjectMapper#writeValueAsBytes(Object)} and
     * {@link ObjectMapper#writeValueAsString(Object)} seem to have issues with writing
     * a raw string into a new string, resulting in obsolete quotation marks at begin and end of the actual string.
     * Therefore, this method contains a dedicated logic part for dealing with raw strings.
     * <p>
     * Also, we have to do a little workaround with {@link String#getBytes(Charset)} due to OkHttp enforcing a standard
     * value for the header "Content-Type" in case of raw strings. See
     * <a href="https://github.com/square/okhttp/issues/2099">https://github.com/square/okhttp/issues/2099</a>
     * for more information on that issue.
     */
    <T> RequestBody createRequestBody(Map<String, String> headers, T httpPayload) throws JsonProcessingException {

        byte[] body;
        if (httpPayload.getClass().equals(String.class)) {
            body = ((String) httpPayload).getBytes(StandardCharsets.UTF_8);
        } else {
            body = this.objectMapper.writeValueAsBytes(httpPayload);
        }

        if (headers.containsKey(HttpHeaders.CONTENT_TYPE)) {
            return RequestBody.create(body, MediaType.parse(headers.get(HttpHeaders.CONTENT_TYPE)));
        }

        return RequestBody.create(body, MediaType.get(MimeTypes.APPLICATION_JSON));
    }

    <R> HttpResponse<R> evaluateStatusAndRespond(Response response, Class<R> responseType) throws IOException {

        if (this.throwExceptionsOnFailure && !response.isSuccessful()) {
            String message = response.body() != null ? response.body().string() : "HTTP call failed.";
            throw new RestApiException(message, response.code());
        }
        ResponseBody responseBody = response.body();
        if (responseBody == null) {
            throw new RestApiException("Response body is null.");
        }
        if (responseType.equals(Void.class)) {
            return new HttpResponse<>(response.code(), null);
        }
        if (responseType.equals(String.class)) {
            return new HttpResponse<>(response.code(), (R) responseBody.string());
        }

        return new HttpResponse<>(response.code(), this.objectMapper.readValue(responseBody.string(), responseType));
    }
}
